import java.io.*;
import java.util.*;

public class Benchmark {
    String name;
    String logic;
    String status;
    String notes;
    Expr[] annotations;
    Expr[] extrasorts;
    Expr[] extrafuns;
    Expr[] extrapreds;
    Expr[] assumptions;
    Expr formula;

    public static String SAT = "sat";
    public static String UNSAT = "unsat";
    public static String UNKNOWN = "unknown";

    protected void hist_decls(Expr[] decls, Vector hist) {
	for (int i = 0; i < decls.length; i++) {
	    Expr decl = decls[i];
	    int arity = decl.kids.length - 1;
	    if (hist.size() < arity+1)
		hist.setSize(arity+1);
	    Integer j = (Integer)hist.get(arity);
	    j = new Integer(j == null ? 1 : j.intValue()+1);
	    hist.set(arity,j);
	}
    }

    protected void print_hist(PrintStream w, String what, Vector hist) {
	if (hist.size() > 0)
	    w.println(what);
	for (int i = 0, iend = hist.size(); i < iend; i++)
	    if (hist.get(i) != null)
		w.println("  arity "+(new Integer(i)) + ": " 
			  + hist.get(i));
    }

    /* we assume the benchmark has already been scrambled, so we don't
       have to worry about annotations. */
    protected void printStatistics(PrintStream w) {
	if (extrasorts.length > 0)
	    w.println("extrasorts: "+(new Integer(extrasorts.length)));

	Vector histfuns = new Vector();
	hist_decls(extrafuns,histfuns);
	print_hist(w,"extrafuns:",histfuns);

	Vector histpreds = new Vector();
	hist_decls(extrapreds,histpreds);
	print_hist(w,"extrapreds:",histpreds);

	HashMap histops = new HashMap();
	for (int i = 0; i < assumptions.length; i++)
	    assumptions[i].hist_ops(histops);
	formula.hist_ops(histops);

	w.println("operators:");
	Iterator i = histops.keySet().iterator();
	while (i.hasNext()) {
	    String op = (String)i.next();
	    if (op.equals(Expr.INDEXED_ID) 
		|| op.equals(Expr.ID)
		|| op.equals(Expr.FVAR)
		|| op.equals(Expr.VAR) 
		|| op.equals(Expr.ATOM) 
		|| op.equals(Expr.CONST) 
		|| op.equals(Expr.INDEXED_CONST)
		|| op.equals(Expr.NUMBER) 
		|| op.equals(Expr.DECL))
		continue;
	    w.println("  "+op+": "+histops.get(op));
	}
    }

    protected void print_decls(PrintStream w, String what, Expr[] decls) {
	if (decls.length > 0) {
	    w.print(what);
	    w.print("(");
	    for (int i = 0; i < decls.length; i++) {
		decls[i].print(w);
		w.println("");
	    }
	    w.println(")");
	}
    }

    public void print(PrintStream w) {
	w.print("(benchmark ");
	w.println(name);
	w.print(":logic ");
	w.println(logic);
	w.print(":status ");
	w.println(status);
	if (notes != null) {
	    w.print(":notes \"");
	    w.print(notes);
	    w.println("\"");
	}

	for (int i = 0; i < annotations.length; i++) {
	    annotations[i].print(w);
	    w.println("");
	}
	
	print_decls(w, ":extrasorts ", extrasorts);
	print_decls(w, ":extrafuns ", extrafuns);
	print_decls(w, ":extrapreds ", extrapreds);

	for (int i = 0; i < assumptions.length; i++) {
	    w.print(":assumption ");
	    assumptions[i].print(w);
	    w.println("");
	}

	w.print(":formula ");
	formula.print(w);
	w.println(")");
    }
 
    void scramble_decls(ScrambleData d, Expr[] decls) {
	for (int i = 0, iend = decls.length; i < iend; i++)
	    decls[i].scramble(d);
    }

    void scramble(ScrambleData d) {
	name = "name_removed";
	status = "unknown";
	notes = null;
	annotations = new Expr[0];
	scramble_decls(d,extrasorts);
	scramble_decls(d,extrafuns);
	scramble_decls(d,extrapreds);
	scramble_decls(d,assumptions);
	formula.scramble(d);
    }
}