import java.io.*;
import java.util.*;

public class Expr {

    public Position pos;
    public String op;
    public Expr[] kids;
    public Object[] data;

    public static String VAR = "V";
    public static String FVAR = "FV";
    public static String ID = "VI";
    public static String INDEXED_ID = "II";
    public static String INDEXED_CONST = "IC";
    public static String CONST = "CONST";
    public static String ANNO = "AN";
    public static String PAT = "PAT";
    public static String NOPAT = "NOPAT";
    public static String ATOM = "AT";
    public static String DECL = "DECL";
    public static String NUMBER = "NUMBER";
    public static String NUM_INDICES = "NUM_INDICES";

    public Expr(Position pos, String op, Expr[] kids, Object[] data) {
	this.pos = pos;
	this.op = op;
	this.kids = kids;
	this.data = data;
    }

    public Expr(Position pos, String op) {
	this.pos = pos;
	this.op = op;
	this.kids = new Expr[0];
	this.data = new Object[0];
    }

    public Expr(Position pos, String op, Expr[] kids) {
	this.pos = pos;
	this.op = op;
	this.kids = kids;
	this.data = new Object[0];
    }

    public Expr(Position pos, String op, Object[] data) {
	this.pos = pos;
	this.op = op;
	this.data = data;
	this.kids = new Expr[0];
    }

    public Expr(Position pos, String op, Object data) {
	this.pos = pos;
	this.op = op;
	Object[] tmp = new Object[1];
	tmp[0] = data;
	this.data = tmp;
	this.kids = new Expr[0];
    }

    public Expr(Position pos, String op, Object data1, Object data2) {
	this.pos = pos;
	this.op = op;
	Object[] tmp = new Object[2];
	tmp[0] = data1;
	tmp[1] = data2;
	this.data = tmp;
	this.kids = new Expr[0];
    }

    public Expr(Position pos, String op, Expr kid1, Expr kid2) {
	this.pos = pos;
	this.op = op;
	this.data = new Object[0];
	Expr[] tmp = new Expr[2];
	tmp[0] = kid1;
	tmp[1] = kid2;
	this.kids = tmp;
    }

    public boolean isPatAnno() {
	return op.equals(PAT) || op.equals(NOPAT);
    }

    public void print(PrintStream w) {
	// a few special cases first
	if (op.charAt(0) == 'I') {
	    w.print(data[0]);
	    w.print("[");
	    boolean first = true;
	    for (int i = 1, iend = data.length; i < iend; i++) {
		if (first)
		    first = false;
		else
		    w.print(":");
		w.print(data[i]);
	    }
	    w.print("]");
	    return;
	}

        if (op.equals(PAT))
	    w.print(":pat { ");
	else if(op.equals(NOPAT))
	    w.print(":nopat { ");
	else if (kids.length > 0)
	    w.print("(");
	if (Character.isLowerCase(op.charAt(0))) {
	    w.print(op);
	    w.print(" ");
	}
	boolean first = true;
	for (int i = 0, iend = kids.length; i < iend; i++) {
	    if (kids[i] != null) {
		if (first)
		    first = false;
		else
		    w.print(" ");
		kids[i].print(w);
	    }
	}
	first = true;
	for (int i = 0, iend = data.length; i < iend; i++) {
	    if (data[i] != null) {
		if (first)
		    first = false;
		else
		    w.print(" ");
		w.print(data[i]);
	    }
	}
        if (op.equals(PAT) || op.equals(NOPAT))
	    w.print(" }");
	else if (kids.length > 0)
	    w.print(")");
    }

    public String toString() {
	ByteArrayOutputStream s = new ByteArrayOutputStream();
	PrintStream w = new PrintStream(s);
	print(w);
	return s.toString();
    }

    public void handleError(String msg) {
	pos.print(System.out);
	System.out.println(": classification error.\n"+msg);
	System.exit(0);
    }

    public void hist_ops(HashMap hist) {
	String s = (op.equals(CONST) || op.equals(INDEXED_CONST)
		    ? (String)data[0] : op);
	Integer i = (Integer)hist.get(s);
	i = new Integer(i == null ? 1 : i.intValue() + 1);
	hist.put(s,i);
	for (int j = 0, jend = kids.length; j < jend; j++)
	    if (kids[j] != null) 
		kids[j].hist_ops(hist);
    }

    public void scramble(ScrambleData d) {
	if (op.equals(INDEXED_ID) || op.equals(ID)) {
	    data[0] = "x" + (new Integer(d.varnum++)).toString();
	    return;
	}
	if (op.equals(FVAR)) {
	    data[0] = "$x" + (new Integer(d.varnum++)).toString();
	    return;
	}
	if (op.equals(VAR)) {
	    data[0] = "?x" + (new Integer(d.varnum++)).toString();
	    return;
	}
	Vector kids_left = new Vector();
	for (int i = 0, iend = kids.length; i < iend; i++)
	    if (kids[i] != null) {
		if (kids[i].op.equals(ANNO) || (!d.leave_pats && kids[i].isPatAnno()))
		    kids[i] = null;
		else {
		    kids[i].scramble(d);
		    kids_left.add(kids[i]);
		}
	    }
	boolean is_atom = op.equals(ATOM);
	int iend = kids_left.size();
	if (iend == 1 && is_atom) {
	    /* the issue here is that terms of the form "(<base_term>)" 
	       are not allowed.  So if we have an atom with one child,
	       that means we erased some annotations, and now must
	       just use the <base_term> (the child) instead of 
	       the expression of the form "(<base_term>)". */
	    overwriteWith((Expr)kids_left.elementAt(0));
	    return;
	}

	Expr[] nkids = new Expr[iend];
	int cur = 0;
	for (int j = 0, jend = kids.length; j < jend; j++)
	    if (kids[j] != null)
		nkids[cur++] = kids[j];
	kids = nkids;    

	// set an offset so we can skip kid[0] if this is an ATOM
	int offset;
	if (d.conn_ac.contains(op)) 
	    offset = 0;
	else if (is_atom && d.symb_ac.contains(kids[0])) 
	    offset = 1;
	else
	    // no swapping to do
	    return;

	for (int k = 0, kend = (int)(ScrambleData.FRAC_SWAPS * iend); k < kend; k++) {
	    int a = offset + d.r.nextInt(iend-offset);
	    int b = offset + d.r.nextInt(iend-offset);
	    Expr tmp = kids[a];
	    kids[a] = kids[b];
	    kids[b] = tmp;
	    if (debug_scrambled_args) 
		System.err.println("scrambled " + op + " at " 
				   + pos.toString());
	}
    }

    protected static boolean debug_scrambled_args = false;

    protected void overwriteWith(Expr e) {
	pos = e.pos;
	op = e.op;
	kids = e.kids;
	data = e.data;
    }
}
