public class Main {
    
    public Main() { }

    public static void main(String[] args) 
    {
	String input_file = null;
        Parser P = new Parser();       
	Context ctxt = new Context();

	Expr eadd = ctxt.newConst("+");
	Expr emul = ctxt.newConst("*");
	Expr ebvand = ctxt.newConst("bvand");
	Expr ebvnand = ctxt.newConst("bvnand");
	Expr ebvor = ctxt.newConst("bvor");
	Expr ebvnor = ctxt.newConst("bvnor");
	Expr ebvcomp = ctxt.newConst("bvcomp");
	Expr ebvadd = ctxt.newConst("bvadd");
	Expr ebvmul = ctxt.newConst("bvmul");

	P.setContext(ctxt);

	int argix = 0;
	boolean leave_pats = false;
	if (args.length > 0 && args[0].equals("+pats")) {
	    leave_pats = true;
	    argix = 1;
	}

	if (args.length > argix + 2 || (args.length > argix && args[argix].equals("--help"))) {
	    System.err.println("SMT Benchmark Scrambler for SMT-LIB format version 1.2.");
	    System.err.println("Command lines: 1. scrambler [+pats] <seed> <file>");
	    System.err.println("               2. scrambler [+pats] <seed>");
	    System.err.println("               3. scrambler [+pats]");
	    System.err.println("");
	    System.err.println("If the seed is 0, no scrambling will be performed.");
	    System.err.println("If the file is omitted, read from stdin.");
	    System.err.println("+pats leaves :pat and :nopat annotations in place.");
	    System.err.println("With no arguments, read from stdin but output nothing except statistics.");
	    System.exit(0);
	}

	boolean read_seed = false;
	int seed = 0;
	if (args.length > argix) {
	    String seed_str = args[argix];
	    try {
		seed = Integer.decode(seed_str).intValue();
	    }
	    catch (NumberFormatException e) {
		System.err.println("Expected numeric seed as first argument.");
		System.exit(1);
	    }
	    read_seed = true;
	}

	if (args.length == argix + 2)
	    input_file = args[argix + 1];

	try {
	    P.openFile(input_file);
	} 
	catch (Exception e) {
	    System.err.println("Error opening file:\n"+e.toString());
	    System.exit(1);
	}

	Benchmark b = null;
	try {
	    b = P.benchmark();
	} 
	catch (Exception e) {
	    System.err.println("Error reading file:\n"+e.toString());
	    System.exit(1);
	}
	
	ScrambleData d = new ScrambleData(seed, leave_pats);
	d.conn_ac.add("and");
	d.conn_ac.add("or");
	d.conn_ac.add("xor");
	d.conn_ac.add("iff");
	String logic = b.logic;
	if (logic.equals("AUFLIA") ||
	    logic.equals("AUFLIRA") ||
	    logic.equals("AUFNIRA") ||
	    logic.equals("LRA") ||
	    logic.equals("QF_AUFLIA") ||
	    logic.equals("QF_IDL") ||
	    logic.equals("QF_LIA") || 
	    logic.equals("QF_LRA") || 
	    logic.equals("QF_NIA") || 
	    logic.equals("QF_RDL") ||
	    logic.equals("QF_UFIDL") ||
	    logic.equals("QF_UFLIA") ||
	    logic.equals("QF_UFLRA") ||
	    logic.equals("QF_UFNRA") ||
	    logic.equals("UFNIA")) {
	    d.symb_ac.add(eadd);
	    d.symb_ac.add(emul);
	}
	if (logic.equals("QF_AUFBV") ||
	    logic.equals("QF_BV") ||
	    logic.equals("QF_UFBV") ||
	    logic.equals("QF_UFBV32")) {
	    d.symb_ac.add(ebvand);
	    d.symb_ac.add(ebvnand);
	    d.symb_ac.add(ebvor);
	    d.symb_ac.add(ebvnor);
	    d.symb_ac.add(ebvcomp);
	    d.symb_ac.add(ebvadd);
	    d.symb_ac.add(ebvmul);
	}

	if (seed != 0) {
	    b.scramble(d);
	}

	if (read_seed) {
	    b.print(ctxt.w);
	    ctxt.w.println("");
	    ctxt.w.flush();
	}
	else {
	    b.scramble(d);
	    b.printStatistics(ctxt.w);
	    ctxt.w.flush();
	}
    }
    
}
