%{
(* This program is free software.  See LICENSE for more information *)

let parse_error msg =
print_string msg

open Smt_helper




%}
%token <string> NUMERAL_TOK
%token <string> VAR_TOK
%token <string> SYM_TOK
%token <string> STRING_TOK
%token <string> AR_SYMB
%token <string> USER_VAL_TOK
%token TRUE_TOK
%token FALSE_TOK
%token ITE_TOK
%token NOT_TOK
%token IMPLIES_TOK
%token IF_THEN_ELSE_TOK
%token AND_TOK
%token OR_TOK
%token XOR_TOK
%token IFF_TOK
%token EXISTS_TOK
%token FORALL_TOK
%token LET_TOK
%token FLET_TOK
%token NOTES_TOK
%token SORTS_TOK
%token FUNS_TOK
%token PREDS_TOK
%token EXTENSIONS_TOK
%token DEFINITION_TOK
%token AXIOMS_TOK
%token LOGIC_TOK
%token COLON_TOK
%token LPAREN_TOK
%token RPAREN_TOK
%token SAT_TOK
%token UNSAT_TOK
%token UNKNOWN_TOK
%token ASSUMPTION_TOK
%token FORMULA_TOK
%token STATUS_TOK
%token BENCHMARK_TOK
%token EXTRASORTS_TOK
%token EXTRAFUNS_TOK
%token EXTRAPREDS_TOK
%token LANGUAGE_TOK
%token DOLLAR_TOK
%token EQUALS_TOK
%token DISTINCT_TOK
%token SEMICOLON_TOK
%token EOF_TOK

%start benchmark
%type <Smt_helper.benchmark> benchmark

%%
benchmark:
    LPAREN_TOK BENCHMARK_TOK bench_name bench_attributes RPAREN_TOK
    {
      
	(
      	let b = BENCH($3,$4) in
      	  if not (Hashtbl.mem symbol_table "status") then 
            raise (Validation_error "Status not defined.")
          else 
      	    if not (Hashtbl.mem symbol_table "formula") then 
              raise (Validation_error "Formula not defined.")
            else 
      	      if not (Hashtbl.mem symbol_table "logic") then 
                raise (Validation_error "Logic not defined.")
              else 
              (
		      	Hashtbl.clear symbol_table;
		      	Hashtbl.clear extrasorts;
              	Hashtbl.clear extrafuns;
	    	  	Hashtbl.clear extrapreds;
	      
	      		Hashtbl.replace symbol_table "---" "---";
		      	Hashtbl.replace extrasorts "---" "---";
		      	Hashtbl.replace extrafuns "---" "---";
	    	  	Hashtbl.replace extrapreds "---" "---";
	      	
			Hashtbl.replace extrasorts "Int" "Int";
		      	Hashtbl.replace extrasorts "Real" "Real";
		      	Hashtbl.replace extrasorts "Array" "Array";
				Hashtbl.replace extrapreds "<=" "<=";
				Hashtbl.replace extrapreds ">=" ">=";
				Hashtbl.replace extrapreds ">" ">";
				Hashtbl.replace extrapreds "<" "<";
				Hashtbl.replace extrapreds "select" "select";
				Hashtbl.replace extrafuns "+" "+";
				Hashtbl.replace extrafuns "-" "-";
				Hashtbl.replace extrafuns "*" "*";
				Hashtbl.replace extrafuns "select" "select";
				Hashtbl.replace extrafuns "store" "store";
				Hashtbl.replace extrafuns "~" "~";
                b;
              )
      )
    }

bench_name:
    SYM_TOK 
    { 
      NAME($1)
    }

bench_attributes:
    bench_attribute bench_attributes {$1::$2}
  | bench_attribute {[$1]}

bench_attribute:
    COLON_TOK ASSUMPTION_TOK an_formula {ASSUMPTION($3)}
  | COLON_TOK FORMULA_TOK an_formula 
    {
      if Hashtbl.mem symbol_table "formula" then 
        raise (Validation_error "More than one formula definition found.")
      else (
      	Hashtbl.replace symbol_table "formula" "exists";
      	FORMULA($3) 
      )
    }
  | COLON_TOK STATUS_TOK status 
    {
      if Hashtbl.mem symbol_table "status" then 
        raise (Validation_error "More than one status definition found.")
      else (
      	Hashtbl.replace symbol_table "status" "exists";
      	STATUS($3)
      )
    }
  | COLON_TOK LOGIC_TOK logic_name 
    {
      if Hashtbl.mem symbol_table "logic" then 
        raise (Validation_error "More than one logic definition found.")
      else (
      	Hashtbl.replace symbol_table "logic" "exists";
      	LOGIC($3) 
      )
    }
  | COLON_TOK EXTRASORTS_TOK LPAREN_TOK sort_symb_decls RPAREN_TOK {EXTRASORTS($4)}
  | COLON_TOK EXTRAFUNS_TOK LPAREN_TOK fun_symb_decls RPAREN_TOK {EXTRAFUNS($4)}
  | COLON_TOK EXTRAPREDS_TOK LPAREN_TOK pred_symb_decls RPAREN_TOK {EXTRAPREDS($4)}
  | COLON_TOK NOTES_TOK STRING_TOK {NOTES($3)}
  | annotation {ANNOTATION($1)}

logic_name:
    SYM_TOK {NAME($1)}

status:
    SAT_TOK {SAT}
  | UNSAT_TOK {UNSAT}
  | UNKNOWN_TOK {UNKNOWN}

sort_symbs:
    sort_symb sort_symbs
    { 
      (* returns a list of sort names *)
      $1::$2
    }
  | sort_symb 
    { 
      (* returns a singleton list of one sort name *)
      [$1]
    }

sort_symb_decls:
    sort_symb_decls sort_symb_decl
    { 
      (* returns a list of sort names *)
      $2::$1
    }
  | sort_symb_decl 
    { 
      (* returns a singleton list of one sort name *)
      [$1]
    }


fun_symb_decls:
    fun_symb_decls fun_symb_decl { $2::$1 }
  | fun_symb_decl { [$1] }



fun_symb_decl:
    LPAREN_TOK fun_symb sort_symb sort_symbs annotations RPAREN_TOK
    {
      (* returns the signature of a function with annotations *)
        if Hashtbl.mem extrafuns $2 then
          raise (Validation_error ("function "^$2^" already defined."))
        else (
       	  Hashtbl.add extrafuns $2 $2;
	  FUNC_SIG($2,$4, $3,$5))
    }
  | LPAREN_TOK fun_symb sort_symb sort_symbs RPAREN_TOK
    {
      (* returns the signature of a function without annotations *)
        if Hashtbl.mem extrafuns $2 then
          raise (Validation_error ("function "^$2^" already defined."))
        else (
          Hashtbl.add extrafuns $2 $2;
          FUNC_SIG($2,$4,$3,[]))
    }
  | LPAREN_TOK fun_symb sort_symb annotations RPAREN_TOK
    {
      (* returns the signature of a function with annotations *)
        if Hashtbl.mem extrafuns $2 then
          raise (Validation_error ("function "^$2^" already defined."))
        else (
          Hashtbl.add extrafuns $2 $2;
          FUNC_SIG($2,[],$3,$4))
    }
  | LPAREN_TOK fun_symb sort_symb RPAREN_TOK
    {
      (* returns the signature of a function without annotations *)
        if Hashtbl.mem extrafuns $2 then
          raise (Validation_error ("function "^$2^" already defined."))
        else (
          Hashtbl.add extrafuns $2 $2;
          FUNC_SIG($2,[],$3,[]))
    }

pred_symb_decls:
    pred_symb_decls pred_symb_decl {$2::$1}
  | pred_symb_decl {[$1]}

pred_symb_decl:
    LPAREN_TOK pred_symb sort_symbs annotations RPAREN_TOK
    {
      (* returns the signature of a predicate with annotations *)
        if Hashtbl.mem extrapreds $2 then
          raise (Validation_error ("predicate "^$2^" already defined."))
        else (
	  Hashtbl.add extrapreds $2 $2;
	  PRED_SIG($2,$3,$4))
    }
  | LPAREN_TOK pred_symb annotations RPAREN_TOK
    {
      (* returns the signature of a predicate with annotations *)
      if Hashtbl.mem extrapreds $2 then
        raise (Validation_error ("predicate "^$2^" already defined."))
      else (
      	Hashtbl.add extrapreds $2 $2;
        PRED_SIG($2,[],$3))
    }
  | LPAREN_TOK pred_symb sort_symbs RPAREN_TOK
    {
      (* returns the signature of a predicate without annotations *)
      if Hashtbl.mem extrapreds $2 then
        raise (Validation_error ("predicate "^$2^" already defined."))
      else (
      	Hashtbl.add extrapreds $2 $2;
        PRED_SIG($2,$3,[]))
    }
  | LPAREN_TOK pred_symb  RPAREN_TOK
    {
      (* returns the signature of a predicate without annotations *)
      if Hashtbl.mem extrapreds $2 then
        raise (Validation_error ("predicate "^$2^" already defined."))
      else (
      	Hashtbl.add extrapreds $2 $2;
        PRED_SIG($2,[],[]))
    }

an_formulas:
    an_formula an_formulas {$1::$2}
  | an_formula {[$1]}

an_formula:
    an_atom {$1}
  | LPAREN_TOK connective an_formulas annotations RPAREN_TOK
    {
      (* returns a formula for one of the boolean connectives with annotations *)
      match ($2,$3) with
        (NOT_TOK,[expr1]) -> 
            FA(NOT(expr1),$4)
      | (IMPLIES_TOK, [expr1;expr2]) -> 
            FA(IMPLIES(expr1,expr2),$4)
      | (IF_THEN_ELSE_TOK, [expr1; expr2; expr3]) -> 
            FA(IF_THEN_ELSE(expr1,expr2,expr3),$4)
      | (AND_TOK, exprl) -> 
            FA(AND(exprl),$4)
      | (OR_TOK, exprl) -> 
            FA(OR(exprl),$4)
      | (XOR_TOK, exprl)-> 
            FA(XOR(exprl),$4)
      | (IFF_TOK, exprl)-> 
            FA(IFF(exprl),$4)
      | _ -> raise Connective_rank_mismatch (* means that a connective was used with a number of arguments that doesn't make sense *)
    }
  | LPAREN_TOK connective an_formulas RPAREN_TOK
    {
      (* returns a formula for one of the boolean connectives without annotations *)
      match ($2,$3) with
        (NOT_TOK,[expr1]) -> 
            F(NOT(expr1))
      | (IMPLIES_TOK, [expr1;expr2]) -> 
            F(IMPLIES(expr1,expr2))
      | (IF_THEN_ELSE_TOK, [expr1; expr2; expr3]) -> 
            F(IF_THEN_ELSE(expr1,expr2,expr3))
      | (AND_TOK, exprl) -> 
            F(AND(exprl))
      | (OR_TOK, exprl) -> 
            F(OR(exprl))
      | (XOR_TOK, exprl)-> 
            F(XOR(exprl))
      | (IFF_TOK, exprl)-> 
            F(IFF(exprl))
      | _ -> raise Connective_rank_mismatch (* means that a connective was used with a number of arguments that doesn't make sense *)
      
    }
  | LPAREN_TOK quant_symb quant_vars an_formula annotations RPAREN_TOK
    {
      (* returns a quantified formula with annotations *)
      FA(QUANT($2,$3,$4),$5)
    }
  | LPAREN_TOK quant_symb quant_vars an_formula RPAREN_TOK
    {
      (* returns a quantified formula with no annotations *)
      F(QUANT($2,$3,$4))
    }
  | LPAREN_TOK LET_TOK LPAREN_TOK VAR_TOK an_term RPAREN_TOK an_formula annotations RPAREN_TOK
    {
      (* returns a let formula with annotations *)
      FA(LET($4,$5,$7),$8)
    }
  | LPAREN_TOK LET_TOK LPAREN_TOK VAR_TOK an_term RPAREN_TOK an_formula RPAREN_TOK
    {
      (* returns a let formula with no annotations *)
      F(LET($4,$5,$7))
    }
  | LPAREN_TOK FLET_TOK LPAREN_TOK fvar an_formula RPAREN_TOK an_formula annotations RPAREN_TOK
    {
      (* returns a flet formula with annotations *)
      FA(FLET($4,$5,$7),$8)
    }
  | LPAREN_TOK FLET_TOK LPAREN_TOK fvar an_formula RPAREN_TOK an_formula RPAREN_TOK
    {
      (* returns a flet formula with no annotations *)
      F(FLET($4,$5,$7))
    }

quant_vars:
    quant_vars quant_var 
    {
      (* returns a list of binders *)
      $2::$1
    }
  | quant_var 
    {
      (* returns a singleton list with one binder in it *)
      [$1]
    }

quant_var: 
    LPAREN_TOK VAR_TOK sort_symb RPAREN_TOK
    {
      (* returns the name of the variable along with the name of its sort *)
      ($2,$3)
    }

quant_symb:
    EXISTS_TOK {EXISTS}
  | FORALL_TOK {FORALL}

connective:
    NOT_TOK {NOT_TOK}
  | IMPLIES_TOK {IMPLIES_TOK}
  | IF_THEN_ELSE_TOK {IF_THEN_ELSE_TOK}
  | AND_TOK {AND_TOK}
  | OR_TOK {OR_TOK}
  | XOR_TOK {XOR_TOK}
  | IFF_TOK {IFF_TOK}


an_atom:
    prop_atom 
    {
      (* returns a formula for a propositional atom with no annotations *)
      F($1)
    }
  | LPAREN_TOK prop_atom annotations RPAREN_TOK 
    {
      (* returns a formula for a propositional atom with annotations *)
      FA($2,$3)
    }
  | LPAREN_TOK pred_symb an_terms annotations RPAREN_TOK
    {
      (* returns a formula for a predicate that has 1 or more arguments with annotations*)
      (
        if not (Hashtbl.mem extrapreds $2) then
          raise (Validation_error ("predicate "^$2^" not yet defined."));
        FA(PRED($2,$3),$4)
      )
    }
  | LPAREN_TOK pred_symb an_terms RPAREN_TOK
    {
      (* returns a formula for a predicate that has 1 or more arguments with no annotations*)
      (
        if not (Hashtbl.mem extrapreds $2) then
          raise (Validation_error ("predicate "^$2^" not yet defined."));
        F(PRED($2,$3))
      )
    }
  | LPAREN_TOK EQUALS_TOK an_term an_terms annotations RPAREN_TOK
    {
      (* returns a formula for equality with annotations
         Note that it requires at least two terms*)
      FA(EQUALS($3::$4),$5)
    }
  | LPAREN_TOK EQUALS_TOK an_term an_terms RPAREN_TOK
    {
      (* returns a formula for equality with no annotations
         Note that it requires at least two terms*)      
      F(EQUALS($3::$4))
    }
  | LPAREN_TOK DISTINCT_TOK an_term an_terms annotations RPAREN_TOK
    {
      (* returns a formula for distinct with annotations
         Note that it requires at least two terms*)
      FA(DISTINCT($3::$4),$5)
    }
  | LPAREN_TOK DISTINCT_TOK an_term an_terms RPAREN_TOK
    {
      (* returns a formula for distinct with no annotations
         Note that it requires at least two terms*)
      F(DISTINCT($3::$4))
    }


prop_atom:
  | TRUE_TOK
    {
      (* returns a formula for true *)
      TRUE_PRED
    }
  | FALSE_TOK
    { 
      (* returns a formula for false *)
      FALSE_PRED

    }
  | fvar
    {
      (* returns a formula variable that should be bound with a flet formula higher in the parse tree *)
      FVAR($1)
    }
  | pred_symb 
    {
      (* returns a predicate with no arguments (a propositional variable) *)
      (
        if not (Hashtbl.mem extrapreds $1) then
          raise (Validation_error ("predicate "^$1^" not yet defined."));
        PRED($1,[])
      )
    }
  

an_terms:
    an_term an_terms 
    {
      (* returns a list of terms*)
      $1::$2
    }
  | an_term 
    { 
      (* returns a singleton list of terms *)
      [$1]
    }

an_term:
    basic_term 
    { 
      (* returns a basic term with no annotations *)
      T($1)
    }
  | LPAREN_TOK basic_term annotations RPAREN_TOK 
    {
      (* adds annotations to a basic term and returns an annotated term *)
      TA($2,$3)
    }
  | LPAREN_TOK fun_symb an_terms annotations RPAREN_TOK
    {
      (* returns a function term with at least one argument with annotations *)
      (
        if not (Hashtbl.mem extrafuns $2) then
          raise (Validation_error ("function "^$2^" not yet defined."));
        TA(FUN($2,$3),$4)
      )
    }
  | LPAREN_TOK fun_symb an_terms RPAREN_TOK
    {
      (* returns a function term with at least one argument with no annotations *)
      (
        if not (Hashtbl.mem extrafuns $2) then
          raise (Validation_error ("function "^$2^" not yet defined."));
        T(FUN($2,$3))
      )
    }
  | LPAREN_TOK ITE_TOK an_formula an_term an_term annotations RPAREN_TOK
    {
      (* returns an ITE term with annotations *)
      TA(ITE($3,$4,$5),$6)
    }
  | LPAREN_TOK ITE_TOK an_formula an_term an_term RPAREN_TOK
    {
      (* returns an ITE term with no annotations *)
      T(ITE($3,$4,$5))
    }


basic_term:
    VAR_TOK 
    { 
      (* returns a variable term*)
      VAR($1)
            
    }
  | NUMERAL_TOK 
    {
      (* returns a numeral term*)
      NUMERAL(int_of_string($1))
    }
  | fun_symb 
    {
      (* returns a function term with no arguments (a constant)*)
      (
        if not (Hashtbl.mem extrafuns $1) then
          raise (Validation_error ("function "^$1^" not yet defined."));
        FUN($1,[])
      )
    }

annotations:
    annotation annotations 
    { 
      (*  returns a list of annotations*)
      ($1::$2)
    }
  | annotation 
    {
      (* returns a singleton list of annotations *)
      [$1]
    }
  
annotation:
    attribute 
    {
      (* returns the attribute's name and None because it doesn't have a corresponding value *)
      ($1,None)
    }
  | attribute user_value 
    {
      (* returns the attribute's name along with its corresponding value*)
      ($1,Some $2)
    }

user_value:
    USER_VAL_TOK
    {
    	$1
    }

sort_symb_decl:
    SYM_TOK 
    { 
      (* returns the name of the sort symbol*)
      (
        Hashtbl.add extrasorts $1 $1;
        $1 
      )
    }

sort_symb:
    SYM_TOK 
    { 
      (* returns the name of the sort symbol*)
      
      (
        if not (Hashtbl.mem extrasorts $1) then
	  raise (Validation_error ("sort "^$1^" is not defined."));
        $1
      )
    }

pred_symb:
    SYM_TOK 
    {
        (* returns the name of the predicate *)
        $1
    }
  | AR_SYMB 
    {
        (* returns the name of the predicate  *)
        $1
    }

fun_symb:
    SYM_TOK
    {
        (* returns the name of the function *)
        $1
    }
  | AR_SYMB 
    { 
        (* returns the name of the function *)
        $1
    }


attribute:
    COLON_TOK SYM_TOK { ":"^ $2}

fvar:
    DOLLAR_TOK SYM_TOK { "$" ^ $2}
